import sys
import argparse

import laniopy as la
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as ani

AI_RANGE_DEFS = {
    la.LanioAiRange.RANGE_100MV: {'unit': '$\\pm$100mV', 'min': -0.1, 'max': 0.1},
    la.LanioAiRange.RANGE_1V: {'unit': '$\\pm$1V', 'min': -1.0, 'max': 1.0},
    la.LanioAiRange.RANGE_10V: {'unit': '$\\pm$10V', 'min': -10.0, 'max': 10.0},
    la.LanioAiRange.RANGE_30V: {'unit': '$\\pm$30V', 'min': -30.0, 'max': 30.0},
    la.LanioAiRange.RANGE_20MA_250OHM: {'unit': '20mA', 'min': 0.0, 'max': 20.0},
    la.LanioAiRange.RANGE_20MA_50OHM: {'unit': '20mA', 'min': 0.0, 'max': 20.0},
    la.LanioAiRange.RANGE_TC: {'unit': '$^\\circ$C', 'min': -10.0, 'max': 100.0},
}

# プロットを行う時間を指定 [ms]
PLOT_MAX_TIME = 60000
# データ取得間隔を指定 [ms]
PLOT_INTERVAL = 1000
PLOT_NUMS = PLOT_MAX_TIME // PLOT_INTERVAL

# 引数をパース
parser = argparse.ArgumentParser(prog=sys.argv[0], description='LANIO analog input plot')
parser.add_argument('-a', '--address', type=str, required=True)
parser.add_argument('-p', '--port', type=int, required=True)
args = parser.parse_args()

# LANIOに接続
lanio = la.Lanio()
if not lanio.connect(args.address, args.port, 1):
    sys.exit('connect error')

if lanio.getAiNum() < 1:
    lanio.disconnect()
    sys.exit('no Analog Input')

# 図を作成
fig, hostAx = plt.subplots(figsize=(12.8, 7.2))
fig.subplots_adjust(right=0.7)

# プロットエリアを作成
hostAx.get_yaxis().set_visible(False)
hostAx.set_xlabel('Time (s)')
hostAx.set_xlim((0, (PLOT_NUMS - 1) * PLOT_INTERVAL / 1000))
axes = [hostAx.twinx() for _ in range(lanio.getAiNum())]
x = np.linspace(0, (PLOT_NUMS - 1) * PLOT_INTERVAL / 1000, PLOT_NUMS)
ys = np.zeros((lanio.getAiNum(), PLOT_NUMS), dtype=np.float64)

# ラインを作成
lines = []
cmap = plt.get_cmap('tab10')
for i in range (lanio.getAiNum()):
    aiRange = lanio.getAnalogInputRange(i)
    aiRangeDef = AI_RANGE_DEFS[aiRange]
    label = f'AI{i + 1} ({aiRangeDef["unit"]})'

    axes[i].set_ylim((aiRangeDef['min'], aiRangeDef['max']))
    axes[i].set_ylabel(label)
    axes[i].spines['right'].set_position(('axes', 1 + (i * 0.1)))

    line, = axes[i].plot([], [], color=cmap(i), label=label)
    lines.append(line)
# 最前面に凡例を表示
axes[lanio.getAiNum() - 1].legend(lines, [line.get_label() for line in lines])

# 更新用関数
def update(frameNum):
    for i in range(lanio.getAiNum()):
        _, value = lanio.getAnalogInputByValue(i)
        ys[i][frameNum] = value
        lines[i].set_data(x[:frameNum + 1], ys[i][:frameNum + 1])
    return lines

# アニメーショングラフを表示
a = ani.FuncAnimation(fig, update, frames=PLOT_NUMS, interval=PLOT_INTERVAL, repeat=False)
plt.show()
