from .error import LanioError, LanioErrorType
from .model import LanioModel, LanioDeviceType, LANIO_SPEC_DEFS
from .param import LanioAiRange, LanioAiAdcSps, LanioAiTcType, LanioAiTransmitCycle, LanioTcTransmitRange, LanioAoRange

import abc
import socket
import select
import time
from typing import Union, Tuple, List


class LanioConnection:
    def __init__(self) -> None:
        self.__sock: Union[socket.socket, None] = None

    def clearBuffer(self) -> None:
        if self.__sock is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        try:
            rfd = self.__sock.fileno()
            rlist, *_ = select.select([rfd], [], [], 0.01)
            if rfd in rlist:
                res = self.__sock.recv(4096)
                if res == b'':
                    # socket broken
                    raise LanioError(LanioErrorType.INVALIDRETURNVALUE)
        except (select.error, OSError):
            pass

    def sendCommand(self, command: Union[bytes, bytearray]) -> None:
        if self.__sock is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        total = 0
        while total < len(command):
            sent = self.__sock.send(command[total:])
            if sent == 0:
                # socket broken
                raise LanioError(LanioErrorType.INVALIDOPERATION)
            total += sent

    def recieveResponseByLength(self, length: int) -> bytes:
        if self.__sock is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        total = 0
        response = bytearray()
        while total < length:
            chunk = self.__sock.recv(length - total)
            if chunk == b'':
                # socket broken
                raise LanioError(LanioErrorType.INVALIDRETURNVALUE)
            total += len(chunk)
            response.extend(chunk)
        return response

    def recieveResponseUntilBytes(self, until: bytes) -> bytes:
        MAX_READSIZE = 128
        if self.__sock is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        response = bytearray()
        while True:
            chunk = self.__sock.recv(MAX_READSIZE, socket.MSG_PEEK)
            if chunk == b'':
                # socket broken
                raise LanioError(LanioErrorType.INVALIDRETURNVALUE)
            pos = chunk.find(until)
            if pos == -1:
                response.extend(self.__sock.recv(MAX_READSIZE))
            else:
                response.extend(self.__sock.recv(pos + 1))
                break
        return response

    def transmitCommand(self, command: Union[bytes, bytearray], recvLen: int) -> bytes:
        try:
            self.clearBuffer()
            self.sendCommand(command)
            return self.recieveResponseByLength(recvLen)
        except (socket.timeout, TimeoutError):
            raise LanioError(LanioErrorType.TIMEOUT)

    def transmitCommandUntil(self, command: Union[bytes, bytearray], until: bytes) -> bytes:
        try:
            self.clearBuffer()
            self.sendCommand(command)
            return self.recieveResponseUntilBytes(until)
        except (socket.timeout, TimeoutError):
            raise LanioError(LanioErrorType.TIMEOUT)

    def connect(self, host: str, port: int, timeout: float) -> bool:
        if self.__sock is not None:
            raise LanioError(LanioErrorType.ALREADYCONNECTED)
        try:
            addr = socket.gethostbyname(host)
            self.__sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.__sock.settimeout(timeout)
            self.__sock.connect((addr, port))
            return True
        except:
            self.__sock = None
            return False

    def getTimeout(self) -> Union[float, None]:
        return None if self.__sock is None else self.__sock.gettimeout()

    def setTimeout(self, timeout: Union[float, None]):
        if self.__sock is not None:
            self.__sock.settimeout(timeout)

    def disconnect(self) -> None:
        if self.__sock is not None:
            self.__sock.shutdown(socket.SHUT_RDWR)
            self.__sock.close()
            self.__sock = None


class LanioInterface(abc.ABC):
    def __init__(self) -> None:
        super().__init__()

    @classmethod
    @abc.abstractmethod
    def requestIdModel(cls, conn: LanioConnection) -> Tuple[LanioDeviceType, int, LanioModel]:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    # デジタル入出力操作
    @abc.abstractmethod
    def outPio(self, doNum: int, state: bool) -> bool:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    @abc.abstractmethod
    def outPioAll(self, state: List[bool]) -> bool:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    @abc.abstractmethod
    def inPio(self, diNum: int) -> bool:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    @abc.abstractmethod
    def inPioAll(self) -> List[bool]:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    @abc.abstractmethod
    def getOut(self, doNum: int) -> bool:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    @abc.abstractmethod
    def getOutAll(self) -> List[bool]:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    # アナログ入力操作
    def getAnalogInput(self, aiNum: int) -> int:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    def getAnalogInputAll(self) -> List[int]:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    def getAnalogInputByValue(self, aiNum: int) -> Tuple[LanioAiRange, float]:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    def setAnalogInputRange(self, aiNum: int, aiRange: LanioAiRange) -> bool:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    def getAnalogInputRange(self, aiNum: int) -> LanioAiRange:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    def setAnalogInputAdcSps(self, sps: LanioAiAdcSps) -> bool:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    def setAnalogThresholds(self, ioNum: int, outBlink: bool, thresholdLow: int, thresholdHigh: int) -> bool:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    def getAnalogThresholds(self, ioNum: int) -> Tuple[bool, int, int]:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    def setTcConfig(self, aiNum: int, tcType: LanioAiTcType, coldJunctionComp: bool, burnOutDetect: bool, burnOutValue: bool) -> bool:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    def getTcConfig(self, aiNum: int) -> Tuple[LanioAiTcType, bool, bool, int]:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    def setTcType(self, aiNum: int, tcType: LanioAiTcType) -> bool:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    # アナログ出力
    def setAnalogOutput(self, aoNum: int, aoRange: LanioAoRange, val: int) -> bool:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    def setAnalogOutputMulti(self, aoNums: List[int], aoRanges: List[LanioAoRange], vals: List[int]) -> bool:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    def getAnalogOutput(self, aoNum: int) -> Tuple[LanioAoRange, int]:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    def setAnalogOutputVoltage(self, aoNum: int, voltage: float) -> bool:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    def setAnalogOutputCurrent(self, aoNum: int, milliAmpere: float) -> bool:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    def getAnalogOutputByValue(self, aoNum: int) -> Tuple[LanioAoRange, float]:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    def setInitialAnalogOutput(self, aoNum: int, aoRange: LanioAoRange, val: int) -> bool:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    def setInitialAnalogOutputVoltage(self, aoNum: int, voltage: float) -> bool:
        raise LanioError(LanioErrorType.INVALIDOPERATION)

    def setInitialAnalogOutputCurrent(self, aoNum: int, milliAmpere: float) -> bool:
        raise LanioError(LanioErrorType.INVALIDOPERATION)


class LanioDigital(LanioInterface):
    '''
        デジタル入出力モデルの制御コマンド
    '''
    @classmethod
    def requestIdModel(cls, conn: LanioConnection) -> Tuple[LanioDeviceType, int, LanioModel]:
        res = conn.transmitCommand(b'\x55\x55', 2)
        # Old Command
        modelNum = (res[0] >> 4) & 0x07
        devtype = LanioDeviceType.DIGITAL
        id = ~res[0] & 0x0f
        model = LanioModel.NONE
        if (res[1] >> 4) == 0x07:
            # Serial
            if modelNum == 0x02:
                model = LanioModel.LA485R
            elif modelNum == 0x03:
                model = LanioModel.LA485T
            elif modelNum == 0x04:
                model = LanioModel.LA232R
            elif modelNum == 0x05:
                model = LanioModel.LA232T
            if model == LanioModel.LA485R or model == LanioModel.LA485T:
                id = 0
        else:
            # Digital
            model = LanioModel(modelNum)

        if model == LanioModel.NONE:
            raise LanioError(LanioErrorType.NOLANIO)

        return (devtype, id, model)

    def __init__(self, devtype: LanioDeviceType, id: int, model: LanioModel, conn: LanioConnection, connPio: Union[LanioConnection, None] = None) -> None:
        super().__init__()
        self.__id = id
        self.__type = devtype
        self.__model = model
        self.__conn = conn
        self.__connPio = connPio

    def xportSetCp(self, mask: int, value: int) -> int:
        if self.__connPio is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        cmd = bytearray(b'\x1b\x00\x00\x00\x00\x00\x00\x00\x00')
        cmd[1] = mask
        cmd[5] = value
        res = self.__connPio.transmitCommand(cmd, 5)
        if res[0] == 0x1b and res[2:] == b'\x00\x00\x00':
            return res[1]
        else:
            raise LanioError(LanioErrorType.INVALIDRETURNVALUE)

    def xportGetCp(self) -> int:
        if self.__connPio is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        cmd = bytearray(b'\x13\00\00\00\00\00\00\00\00')
        res = self.__connPio.transmitCommand(cmd, 5)
        if res[0] == 0x13 and res[2:] == b'\x00\x00\x00':
            return res[1]
        else:
            raise LanioError(LanioErrorType.INVALIDRETURNVALUE)

    def setOutF0(self, value: int) -> bool:
        cmd = bytearray(b'\xf0\x00')
        cmd[1] = value
        return self.__conn.transmitCommand(cmd, 2) == cmd

    def getOutE0(self) -> int:
        res = self.__conn.transmitCommand(b'\xe0', 2)
        if res[0] == 0xe0:
            return res[1]
        else:
            raise LanioError(LanioErrorType.INVALIDRETURNVALUE)

    def getIn55(self) -> List[bool]:
        res = self.__conn.transmitCommand(b'\x55\x55', 2)
        resModelNum = (res[0] >> 4) & 0x07
        resUnitId = ~res[0] & 0x0f
        if (res[1] >> 4) == 0x07:
            # Serial
            resModel = LanioModel.NONE
            if resModelNum == 0x02:
                resModel = LanioModel.LA485R
            elif resModelNum == 0x03:
                resModel = LanioModel.LA485T
            elif resModelNum == 0x04:
                resModel = LanioModel.LA232R
            elif resModelNum == 0x05:
                resModel = LanioModel.LA232T
            if resModelNum == LanioModel.LA485R or resModelNum == LanioModel.LA485T:
                resUnitId = 0
            if resModel == self.__model and resUnitId == self.__id and (res[1] & 0xf0) == 0x70:
                return [
                    (res[1] & 0x01) == 0x01,  # DI1
                    (res[1] & 0x02) == 0x02,  # DI2
                ]
            else:
                raise LanioError(LanioErrorType.INVALIDRETURNVALUE)
        else:
            # Digital
            if resModelNum == self.__model.value and resUnitId == self.__id and (res[1] & 0xf0) == 0xf0:
                return [
                    (res[0] & 0x80) == 0x80,  # DI1
                    (res[1] & 0x01) == 0x01,  # DI2
                    (res[1] & 0x02) == 0x02,  # DI3
                    (res[1] & 0x04) == 0x04,  # DI4
                    (res[1] & 0x08) == 0x08,  # DI5
                ]
            else:
                raise LanioError(LanioErrorType.INVALIDRETURNVALUE)

    def getInEC(self) -> List[bool]:
        res = self.__conn.transmitCommand(b'\xec', 2)
        if res[0] == 0xec and (res[1] >> 7) == 0:
            return [(res[i] & (1 << i)) == (1 << i) for i in range(7)]
        else:
            raise LanioError(LanioErrorType.INVALIDRETURNVALUE)

    def outPio(self, doNum: int, state: bool) -> bool:
        if self.__model == LanioModel.LA3R2P and self.__type != LanioDeviceType.WIFI:
            mask = 1 << doNum
            value = (1 if state else 0) << doNum
            return (self.xportSetCp(mask, value) & mask) == value
        else:
            do = self.getOutE0()
            do = (do & ~(1 << doNum)) | ((1 if state else 0) << doNum)
            return self.setOutF0(do)

    def outPioAll(self, doStates: List[bool]) -> bool:
        do = 0
        for i in range(len(doStates)):
            do |= (1 if doStates[i] else 0) << i
        if self.__model == LanioModel.LA3R2P and self.__type != LanioDeviceType.WIFI:
            return self.xportSetCp(0x07, do) == do
        else:
            return self.setOutF0(do)

    def inPio(self, diNum: int) -> bool:
        return (self.inPioAll())[diNum]

    def inPioAll(self) -> List[bool]:
        diNum = LANIO_SPEC_DEFS[self.__model].diNum
        if self.__model == LanioModel.LA7P:
            if self.__type == LanioDeviceType.WIFI:
                return self.getInEC()
            else:
                di = self.getIn55()
                diCp = self.xportGetCp()
                di.append((diCp & 0x01) == 0x01)  # Append DI6
                di.append((diCp & 0x02) == 0x02)  # Append DI7
                return di
        else:
            return (self.getIn55())[:diNum]

    def getOut(self, doNum: int) -> bool:
        return (self.getOutAll())[doNum]

    def getOutAll(self) -> List[bool]:
        do = 0
        if self.__model == LanioModel.LA3R2P and self.__type != LanioDeviceType.WIFI:
            do = self.xportGetCp()
        else:
            do = self.getOutE0()
        return [(do & (1 << i)) == (1 << i) for i in range(LANIO_SPEC_DEFS[self.__model].doNum)]


class LanioNew(LanioInterface):
    '''
        新しい体系の制御コマンド
    '''
    @classmethod
    def requestIdModel(cls, conn: LanioConnection) -> Tuple[LanioDeviceType, int, LanioModel]:
       # New Command
        res = conn.transmitCommandUntil(b'MI\xc8', b'\xc8')
        # Validate
        if res[:2] != bytes('mi', 'ascii') or res[-1] != 0xc8:
            # Error
            raise LanioError(LanioErrorType.INVALIDRETURNVALUE)
        # Ok
        devtype = LanioDeviceType.NONE
        unitId = ~res[2] & 0x0f
        model = LanioModel.NONE
        if (res[3] & 0x0f) == 0x0f:
            devtype = LanioDeviceType.N_8P
            # LANIO8 or LANIO4
            modelNum = res[4] & 0x0f
            if modelNum == 0x03:
                model = LanioModel.LA8R
            elif modelNum == 0x04:
                model = LanioModel.LA8T
            elif modelNum == 0x05:
                model = LanioModel.LA8PP
            elif modelNum == 0x06:
                model = LanioModel.LA4T4SP
            elif modelNum == 0x09:
                model = LanioModel.LAN2R2P
        else:
            devtype = LanioDeviceType.ANALOG
            # Analog
            modelNum = res[3] & 0x0f
            if modelNum == 0x08:
                model = LanioModel.LA2R3A
            elif modelNum == 0x09:
                model = LanioModel.LA2A3PP
            elif modelNum == 0x0a:
                model = LanioModel.LA2R3AV2
            elif modelNum == 0x0b:
                model = LanioModel.LA3A2PP
            elif modelNum == 0x0c:
                model = LanioModel.LA5AI
        if model == LanioModel.NONE:
            raise LanioError(LanioErrorType.NOLANIO)

        return (devtype, unitId, model)

    def __init__(self, model: LanioModel, conn: LanioConnection) -> None:
        super().__init__()
        self.__model = model
        self.__conn = conn

    def validateResponse(self, command: str, response: Union[bytes, bytearray]) -> bool:
        return response[:2] == bytes(command.lower(), 'ascii') and response[-1] == 0xc8

    def getInDI(self) -> List[bool]:
        res = self.__conn.transmitCommandUntil(b'DI\xc8', b'\xc8')
        if not self.validateResponse('DI', res):
            raise LanioError(LanioErrorType.INVALIDRETURNVALUE)
        return [(res[2:4][i // 4] & (1 << (i % 4))) != 0 for i in range(LANIO_SPEC_DEFS[self.__model].diNum)]

    def setOutDO(self, states: List[bool]) -> bool:
        body = bytearray(b'\x20\x20')
        for doNum, state in enumerate(states):
            body[doNum // 4] |= (1 if state else 0) << (doNum % 4)
        res = self.__conn.transmitCommandUntil(b'DO' + body + b'\xc8', b'\xc8')
        return self.validateResponse('DO', res)

    def getOutDY(self) -> List[bool]:
        res = self.__conn.transmitCommandUntil(b'DY\xc8', b'\xc8')
        if not self.validateResponse('DY', res):
            raise LanioError(LanioErrorType.INVALIDRETURNVALUE)
        return [(res[2:4][i // 4] & (1 << (i % 4))) != 0 for i in range(LANIO_SPEC_DEFS[self.__model].doNum)]

    def setOutDK(self, doNums: List[int], states: List[bool]) -> bool:
        body = bytearray(b'\x20\x20\x20\x20')
        for doNum, state in zip(doNums, states):
            index = (doNum // 4) * 2
            pos = doNum % 4
            body[index] |= state << pos
            body[index + 1] |= 1 << pos
        res = self.__conn.transmitCommandUntil(b'DK' + body + b'\xc8', b'\xc8')
        return self.validateResponse('DK', res)

    def outPio(self, doNum: int, state: bool) -> bool:
        return self.setOutDK([doNum], [state])

    def outPioAll(self, doStates: List[bool]) -> bool:
        return self.setOutDO(doStates)

    def inPio(self, diNum: int) -> bool:
        return (self.inPioAll())[diNum]

    def inPioAll(self) -> List[bool]:
        return self.getInDI()

    def getOut(self, doNum: int) -> bool:
        return (self.getOutAll())[doNum]

    def getOutAll(self) -> List[bool]:
        return self.getOutDY()


class LanioAnalog(LanioNew):
    '''
        新しい体系の制御コマンド (アナログ)
    '''
    @classmethod
    def requestIdModel(cls, conn: LanioConnection) -> Tuple[LanioDeviceType, int, LanioModel]:
        return super().requestIdModel(conn)

    @classmethod
    def toHex(cls, num: int, bit: int) -> str:
        mask = (1 << bit) - 1
        return f'{num & mask:0{bit // 4}X}'

    @classmethod
    def toInt(cls, val: str, bit: int, signed=True) -> int:
        try:
            mask = (1 << bit) - 1
            num = int(val, 16)
            if signed and num > (mask // 2):
                num = -(~(num - 1) & mask)
            return num
        except ValueError:
            raise LanioError(LanioErrorType.INVALIDRETURNVALUE)

    def __init__(self, model: LanioModel, conn: LanioConnection):
        super().__init__(model, conn)
        self.__conn = conn
        self.__model = model

    # デジタル入出力互換操作
    def outPio(self, doNum: int, state: bool) -> bool:
        # アナログモデルはDKコマンドを受け付けていないため
        out = self.getOutDY()
        out[doNum] = state
        return self.setOutDO(out)

    # アナログ入力
    def getInAI(self, ch: int) -> List[int]:
        cmd = bytearray(b'AI\x00\xc8')
        cmd[2] = 0x20 | (ch & 0x07)
        res = self.__conn.transmitCommandUntil(cmd, b'\xc8')
        if not self.validateResponse('AI', res):
            raise LanioError(LanioErrorType.INVALIDRETURNVALUE)
        # ch が 0x07 のとき一括取得
        aiNum = LANIO_SPEC_DEFS[self.__model].aiNum if ch == 0x07 else 1
        if len(res[3:-1]) != (6 * aiNum):
            raise LanioError(LanioErrorType.INVALIDRETURNVALUE)
        return [
            LanioAnalog.toInt(str(res[3 + 6 * i:3 + 6 * (i + 1)], 'ascii'), 24) for i in range(aiNum)
        ]

    def getRangeCycleAJ(self, aiNum: int) -> Tuple[LanioAiRange, LanioAiTransmitCycle]:
        cmd = bytearray(b'AJ\x00\xc8')
        cmd[2] = 0x20 | (aiNum & 0x07)
        res = self.__conn.transmitCommandUntil(cmd, b'\xc8')
        if not self.validateResponse('AJ', res) or not res[2] == cmd[2]:
            raise LanioError(LanioErrorType.INVALIDRETURNVALUE)
        rangeNum = res[3] & 0x07
        cycleNum = res[4] & 0x07
        return (LanioAiRange(rangeNum), LanioAiTransmitCycle(cycleNum))

    def setRangeAR(self, aiNum: int, aiRange: LanioAiRange) -> bool:
        if self.__model == LanioModel.LA2R3A and aiRange == LanioAiRange.RANGE_TC:
            return False
        cmd = bytearray(b'AR\x00\x00\xc8')
        cmd[2] = 0x20 | (aiNum & 0x07)
        cmd[3] = 0x20 | (aiRange.value & 0x07)
        res = self.__conn.transmitCommandUntil(cmd, b'\xc8')
        return self.validateResponse('AR', res) and res[2] == cmd[2] and res[3] == cmd[3]

    def setSpsAP(self, sps: LanioAiAdcSps) -> bool:
        cmd = bytearray(b'AP\x00\xc8')
        cmd[2] = 0x20 | (sps.value & 0x07)
        res = self.__conn.transmitCommandUntil(cmd, b'\xc8')
        return self.validateResponse('AP', res) and res[2] == cmd[2]

    def setLedMailThreshAL(self, ch: int, outBlink: bool, mailNum: int, thresholdLow: int, thresholdHigh: int) -> bool:
        cmd = bytearray(b'AL\x00\x00\xc8')
        cmd[2] = 0x20 | (ch & 0x07)
        cmd[3] = 0x20 | (0x08 if outBlink else 0x00) | (mailNum & 0x03)
        cmd[4:4] = bytes(LanioAnalog.toHex(thresholdLow, 16), 'ascii')
        cmd[8:8] = bytes(LanioAnalog.toHex(thresholdHigh, 16), 'ascii')
        res = self.__conn.transmitCommandUntil(cmd, b'\xc8')
        return self.validateResponse('AL', res) and res[2:12] == cmd[2:12]

    def getLedMailThreshAK(self, ch: int) -> Tuple[bool, int, int, int]:
        cmd = bytearray(b'AK\x00\xc8')
        cmd[2] = 0x20 | (ch & 0x07)
        res = self.__conn.transmitCommandUntil(cmd, b'\xc8')
        if not self.validateResponse('AK', res) or res[2] != cmd[2]:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        outBlink = (res[3] & 0x08) == 0x08
        mailNum = res[3] & 0x03
        thresholdLow = LanioAnalog.toInt(str(res[4:8], 'ascii'), 16)
        thresholdHigh = LanioAnalog.toInt(str(res[8:12], 'ascii'), 16)
        return (outBlink, mailNum, thresholdLow, thresholdHigh)

    def setTcTC(self, aiNum: int, tcType: LanioAiTcType, coldJunctionComp: bool, burnOutDetect: bool, burnOutValue: bool) -> bool:
        if self.__model == LanioModel.LA2R3A:
            return False
        cmd = bytearray(b'TC\x00\x00\x00\xc8')
        cmd[2] = 0x20 | (aiNum & 0x07)
        cmd[3] = 0x20 | (tcType.value & 0x0f)
        cmd[4] = 0x20 | \
            (0x04 if burnOutValue else 0x00) | \
            (0x02 if burnOutDetect else 0x00) | \
            (0x01 if coldJunctionComp else 0x00)
        res = self.__conn.transmitCommandUntil(cmd, b'\xc8')
        return self.validateResponse('TC', res) and res[2:5] == cmd[2:5]

    def getTcTJ(self, aiNum: int) -> Tuple[LanioAiTcType, bool, bool, bool, LanioTcTransmitRange, int, int]:
        if self.__model == LanioModel.LA2R3A:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        cmd = bytearray(b'TJ\x00\xc8')
        cmd[2] = 0x20 | (aiNum & 0x07)
        res = self.__conn.transmitCommandUntil(cmd, b'\xc8')
        if not self.validateResponse('TJ', res) or res[2] != cmd[2]:
            raise LanioError(LanioErrorType.INVALIDRETURNVALUE)
        tcType = LanioAiTcType(res[3] & 0x0f)
        burnOutValue = (res[4] & 0x04) == 0x04
        burnOutDetect = (res[4] & 0x02) == 0x02
        coldJunctionComp = (res[4] & 0x01) == 0x01
        tcRange = LanioTcTransmitRange(res[5] & 0x03)
        scaleLower = LanioAnalog.toInt(str(res[6:10], 'ascii'), 16)
        scaleUpper = LanioAnalog.toInt(str(res[10:14], 'ascii'), 16)
        return (tcType, burnOutValue, burnOutDetect, coldJunctionComp, tcRange, scaleLower, scaleUpper)

    def getAnalogInput(self, aiNum: int) -> int:
        if aiNum < 0 or LANIO_SPEC_DEFS[self.__model].aiNum <= aiNum:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.getInAI(aiNum)[0]

    def getAnalogInputAll(self) -> List[int]:
        return self.getInAI(0x07)

    def getAnalogInputByValue(self, aiNum: int) -> Tuple[LanioAiRange, float]:
        aiRange = self.getAnalogInputRange(aiNum)
        value = self.getInAI(aiNum)[0]
        measureMax = 0.0
        if aiRange == LanioAiRange.RANGE_100MV:
            measureMax = 0.1
        elif aiRange == LanioAiRange.RANGE_1V:
            measureMax = 1.0
        elif aiRange == LanioAiRange.RANGE_10V:
            measureMax = 10.0
        elif aiRange == LanioAiRange.RANGE_30V:
            measureMax = 30.0
        elif aiRange == LanioAiRange.RANGE_20MA_250OHM or \
                aiRange == LanioAiRange.RANGE_20MA_50OHM:
            measureMax = 20.0
        elif aiRange == LanioAiRange.RANGE_TC:
            measureMax = 3276.8
        else:
            raise LanioError(LanioErrorType.INVALIDRETURNVALUE)
        ret = 0.0
        if value < 0:
            ret = measureMax * value / (1 << 23)
        else:
            ret = measureMax * value / ((1 << 23) - 1)
        return (aiRange, ret)

    def setAnalogInputRange(self, aiNum: int, aiRange: LanioAiRange) -> bool:
        return self.setRangeAR(aiNum, aiRange)

    def getAnalogInputRange(self, aiNum: int) -> LanioAiRange:
        aiRange, _ = self.getRangeCycleAJ(aiNum)
        return aiRange

    def setAnalogInputAdcSps(self, sps: LanioAiAdcSps) -> bool:
        return self.setSpsAP(sps)

    def setAnalogThresholds(self, ioNum: int, outBlink: bool, thresholdLow: int, thresholdHigh: int) -> bool:
        if (0 <= ioNum < LANIO_SPEC_DEFS[self.__model].aiNum or 0 <= ioNum < LANIO_SPEC_DEFS[self.__model].aoNum):
            if (-32768 <= thresholdLow <= 32767 and -32768 <= thresholdHigh <= 32767):
                # メールアラート設定を上書きしないように取得しておく
                settings = self.getLedMailThreshAK(ioNum)
                return self.setLedMailThreshAL(ioNum, outBlink, settings[1], thresholdLow, thresholdHigh)
            else:
                return False
        else:
            return False

    def getAnalogThresholds(self, ioNum: int) -> Tuple[bool, int, int]:
        outBlink, _, thresholdLow, thresholdHigh = self.getLedMailThreshAK(
            ioNum)
        return (outBlink, thresholdLow, thresholdHigh)

    def setTcConfig(self, aiNum: int, tcType: LanioAiTcType, coldJunctionComp: bool, burnOutDetect: bool, burnOutValue: bool) -> bool:
        return self.setTcTC(aiNum, tcType, coldJunctionComp, burnOutDetect, burnOutValue)

    def getTcConfig(self, aiNum: int) -> Tuple[LanioAiTcType, bool, bool, int]:
        tcType, coldJunctionComp, burnOutDetect, burnOutValue, *_ = self.getTcTJ(
            aiNum)
        return (tcType, coldJunctionComp, burnOutDetect, burnOutValue)

    def setTcType(self, aiNum: int, tcType: LanioAiTcType) -> bool:
        return self.setTcConfig(aiNum, tcType, True, True, False)

    # アナログ出力
    def genParamAOAQ(self, aoNums: List[int], aoRanges: List[LanioAoRange], vals: List[int]) -> bytearray:
        param = bytearray(b'')
        for aoNum, aoRange, val in zip(aoNums, aoRanges, vals):
            if aoRange == LanioAoRange.RANGE_20MA:
                val = min(32767, max(val, 0))  # 0x0000 ~ 0x7FFF
            else:
                val = min(32767, max(val, -32768))  # 0x8000 ~ 0x7FFF
            rangeBit = (1 << 3) if aoRange == LanioAoRange.RANGE_20MA else 0
            channelBit = aoNum & 0x07
            param.append(0x20 | rangeBit | channelBit)
            param.extend(bytes(LanioAnalog.toHex(val, 16), 'ascii'))
        param.append(0xc8)
        return param

    def setOutAO(self, aoNums: List[int], aoRanges: List[LanioAoRange], vals: List[int]) -> bool:
        cmd = bytearray(b'AO')
        cmd.extend(self.genParamAOAQ(aoNums, aoRanges, vals))
        res = self.__conn.transmitCommandUntil(cmd, b'\xc8')
        return self.validateResponse('AO', res) and cmd[2:] == res[2:]

    def setInitOutAQ(self, aoNums: List[int], aoRanges: List[LanioAoRange], vals: List[int]) -> bool:
        cmd = bytearray(b'AQ')
        cmd.extend(self.genParamAOAQ(aoNums, aoRanges, vals))
        res = self.__conn.transmitCommandUntil(cmd, b'\xc8')
        return self.validateResponse('AQ', res) and cmd[2:] == res[2:]

    def getOutAY(self, aoNum: int) -> Tuple[LanioAoRange, int]:
        cmd = bytearray(b'AY\x00\xc8')
        cmd[2] = 0x20 | (aoNum & 0x07)
        res = self.__conn.transmitCommandUntil(cmd, b'\xc8')
        if self.validateResponse('AY', res):
            aoRange = LanioAoRange.RANGE_20MA if (
                res[2] & 0x08) != 0 else LanioAoRange.RANGE_10V
            val = LanioAnalog.toInt(str(res[3:7], 'ascii'), 16, True)
            return (aoRange, val)
        else:
            raise LanioError(LanioErrorType.INVALIDRETURNVALUE)

    def setAnalogOutput(self, aoNum: int, aoRange: LanioAoRange, val: int) -> bool:
        return self.setOutAO([aoNum], [aoRange], [val])

    def setAnalogOutputMulti(self, aoNums: List[int], aoRanges: List[LanioAoRange], vals: List[int]) -> bool:
        return self.setOutAO(aoNums, aoRanges, vals)

    def getAnalogOutput(self, aoNum: int) -> Tuple[LanioAoRange, int]:
        return self.getOutAY(aoNum)

    def setAnalogOutputVoltage(self, aoNum: int, voltage: float) -> bool:
        val = 0
        if voltage < 0.0:
            val = int((1 << 15) / 10.0 * voltage - 0.5)  # FFFF ~ 8000
        else:
            val = int(((1 << 15) - 1) / 10.0 * voltage + 0.5)  # 0000 ~ 7FFF
        return self.setOutAO([aoNum], [LanioAoRange.RANGE_10V], [val])

    def setAnalogOutputCurrent(self, aoNum: int, milliAmpere: float) -> bool:
        val = int(((1 << 15) - 1) / 20.0 * milliAmpere + 0.5)  # 0000 ~ 7FFF
        return self.setOutAO([aoNum], [LanioAoRange.RANGE_20MA], [val])

    def getAnalogOutputByValue(self, aoNum: int) -> Tuple[LanioAoRange, float]:
        aoRange, value = self.getOutAY(aoNum)
        if aoRange == LanioAoRange.RANGE_10V:
            measureMax = 10.0
        elif aoRange == LanioAoRange.RANGE_20MA:
            measureMax = 20.0
        ret = 0.0
        if value < 0.0:
            ret = measureMax * value / (1 << 15)
        else:
            ret = measureMax * value / ((1 << 15) - 1)
        return (aoRange, ret)

    def setInitialAnalogOutput(self, aoNum: int, aoRange: LanioAoRange, val: int) -> bool:
        return self.setInitOutAQ([aoNum], [aoRange], [val])

    def setInitialAnalogOutputVoltage(self, aoNum: int, voltage: float) -> bool:
        val = 0
        if voltage < 0.0:
            val = int((1 << 15) / 10.0 * voltage - 0.5)  # FFFF ~ 8000
        else:
            val = int(((1 << 15) - 1) / 10.0 * voltage + 0.5)  # 0000 ~ 7FFF
        return self.setInitOutAQ([aoNum], [LanioAoRange.RANGE_10V], [val])

    def setInitialAnalogOutputCurrent(self, aoNum: int, milliAmpere: float) -> bool:
        val = int(((1 << 15) - 1) / 20.0 * milliAmpere + 0.5)  # 0000 ~ 7FFF
        return self.setInitOutAQ([aoNum], [LanioAoRange.RANGE_20MA], [val])


class Lanio:

    WAIT_FIRST_CONNECTION = 0.1  # [sec]
    TIMEOUT_REQ_MODEL_ID = 1.5  # [sec]

    def __init__(self, devtype: LanioDeviceType = LanioDeviceType.NONE) -> None:
        self.__conn = LanioConnection()
        self.__connPio: LanioConnection | None = None  # ポート番号30704を使用するモデルのみ
        self.__instanceDevType = devtype
        self.__type = devtype
        self.__id: int | None = None
        self.__model = LanioModel.NONE
        self.__io: LanioInterface | None = None

    def connect(self, address: str, port: int, timeout: float, pioPort=30704) -> bool:
        if not self.__conn.connect(address, port, timeout):
            return False
        try:
            time.sleep(self.WAIT_FIRST_CONNECTION)
            self.__type, self.__id, self.__model = self.requestIdModel()
            # モデルに対応した操作クラスを設定
            if self.__type == LanioDeviceType.DIGITAL:
                if self.__model == LanioModel.LA7P or self.__model == LanioModel.LA3R2P:
                    self.__connPio = LanioConnection()
                    self.__connPio.connect(address, pioPort, timeout)
                self.__io = LanioDigital(
                    self.__type, self.__id, self.__model, self.__conn, self.__connPio)
            elif self.__type == LanioDeviceType.WIFI:
                self.__io = LanioDigital(
                    self.__type, self.__id, self.__model, self.__conn)
            elif self.__type == LanioDeviceType.N_8P:
                self.__io = LanioNew(self.__model, self.__conn)
            elif self.__type == LanioDeviceType.ANALOG:
                self.__io = LanioAnalog(self.__model, self.__conn)
            else:
                # 未知のモデル
                self.disconnect()
                return False
            return True
        except LanioError:
            self.disconnect()
            return False

    def disconnect(self) -> None:
        self.__conn.disconnect()
        if self.__connPio is not None:
            self.__connPio.disconnect()
        self.__type = self.__instanceDevType
        self.__id = None
        self.__model = LanioModel.NONE
        self.__io = None

    def getId(self) -> int:
        if self.__id is None:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__id

    def getModel(self) -> LanioModel:
        return self.__model

    def getDoNum(self) -> int:
        return LANIO_SPEC_DEFS[self.__model].doNum

    def getDiNum(self) -> int:
        return LANIO_SPEC_DEFS[self.__model].diNum

    def getAoNum(self) -> int:
        return LANIO_SPEC_DEFS[self.__model].aoNum

    def getAiNum(self) -> int:
        return LANIO_SPEC_DEFS[self.__model].aiNum

    def requestIdModel(self) -> Tuple[LanioDeviceType, int, LanioModel]:
        if self.__type == LanioDeviceType.DIGITAL or self.__type == LanioDeviceType.WIFI:
            # Old Command
            _, unitId, model = LanioDigital.requestIdModel(
                self.__conn)
            return (self.__type, unitId, model)
        elif self.__type == LanioDeviceType.N_8P or self.__type == LanioDeviceType.ANALOG:
            # New Command
            _, unitId, model = LanioNew.requestIdModel(
                self.__conn)
            return (self.__type, unitId, model)
        else:
            # Auto
            prevTimeout = self.__conn.getTimeout()
            self.__conn.setTimeout(self.TIMEOUT_REQ_MODEL_ID)
            ret = (LanioDeviceType.NONE, 0, LanioModel.NONE)
            try:
                # Old Command
                ret = LanioDigital.requestIdModel(self.__conn)
            except LanioError:
                # New Command
                ret = LanioNew.requestIdModel(self.__conn)
            finally:
                self.__conn.setTimeout(prevTimeout)
            return ret

    def outPio(self, doNum: int, state: bool) -> bool:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if doNum < 0 or LANIO_SPEC_DEFS[self.__model].doNum <= doNum:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.outPio(doNum, state)

    def outPioAll(self, doStates: List[bool]) -> bool:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if len(doStates) != LANIO_SPEC_DEFS[self.__model].doNum or \
                LANIO_SPEC_DEFS[self.__model].doNum < 1:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.outPioAll(doStates)

    def inPio(self, diNum: int) -> bool:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if diNum < 0 or LANIO_SPEC_DEFS[self.__model].diNum <= diNum:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.inPio(diNum)

    def inPioAll(self) -> List[bool]:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if LANIO_SPEC_DEFS[self.__model].diNum < 1:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.inPioAll()

    def getOut(self, doNum: int) -> bool:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if doNum < 0 or LANIO_SPEC_DEFS[self.__model].doNum <= doNum:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.getOut(doNum)

    def getOutAll(self) -> List[bool]:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if LANIO_SPEC_DEFS[self.__model].doNum < 1:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.getOutAll()

    def getAnalogInput(self, aiNum: int) -> int:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if aiNum < 0 or LANIO_SPEC_DEFS[self.__model].aiNum <= aiNum:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.getAnalogInput(aiNum)

    def getAnalogInputAll(self) -> List[int]:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if LANIO_SPEC_DEFS[self.__model].aiNum < 1:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.getAnalogInputAll()

    def getAnalogInputByValue(self, aiNum: int) -> Tuple[LanioAiRange, float]:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if aiNum < 0 or LANIO_SPEC_DEFS[self.__model].aiNum <= aiNum:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.getAnalogInputByValue(aiNum)

    def setAnalogInputRange(self, aiNum: int, aiRange: LanioAiRange) -> bool:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if aiNum < 0 or LANIO_SPEC_DEFS[self.__model].aiNum <= aiNum:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.setAnalogInputRange(aiNum, aiRange)

    def getAnalogInputRange(self, aiNum: int) -> LanioAiRange:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if aiNum < 0 or LANIO_SPEC_DEFS[self.__model].aiNum <= aiNum:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.getAnalogInputRange(aiNum)

    def setAnalogInputAdcSps(self, sps: LanioAiAdcSps) -> bool:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if LANIO_SPEC_DEFS[self.__model].aiNum < 1:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.setAnalogInputAdcSps(sps)

    def setAnalogThresholds(self, ioNum: int, outBlink: bool, thresholdLow: int, thresholdHigh: int) -> bool:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if LANIO_SPEC_DEFS[self.__model].aiNum < 1 and LANIO_SPEC_DEFS[self.__model].aoNum < 1:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.setAnalogThresholds(ioNum, outBlink, thresholdLow, thresholdHigh)

    def getAnalogThresholds(self, ioNum: int) -> Tuple[bool, int, int]:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if ioNum < 0 or max(LANIO_SPEC_DEFS[self.__model].aiNum, LANIO_SPEC_DEFS[self.__model].aoNum) <= ioNum:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.getAnalogThresholds(ioNum)

    def setTcConfig(self, aiNum: int, tcType: LanioAiTcType, coldJunctionComp: bool, burnOutDetect: bool, burnOutValue: bool) -> bool:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if aiNum < 0 or LANIO_SPEC_DEFS[self.__model].aiNum <= aiNum:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        if self.__model == LanioModel.LA2R3A:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.setTcConfig(aiNum, tcType, coldJunctionComp, burnOutDetect, burnOutValue)

    def getTcConfig(self, aiNum: int) -> Tuple[LanioAiTcType, bool, bool, int]:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if aiNum < 0 or LANIO_SPEC_DEFS[self.__model].aiNum <= aiNum:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        if self.__model == LanioModel.LA2R3A:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.getTcConfig(aiNum)

    def setTcType(self, aiNum: int, tcType: LanioAiTcType) -> bool:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if aiNum < 0 or LANIO_SPEC_DEFS[self.__model].aiNum <= aiNum:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        if self.__model == LanioModel.LA2R3A:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.setTcType(aiNum, tcType)

    def setAnalogOutput(self, aoNum: int, aoRange: LanioAoRange, val: int) -> bool:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if aoNum < 0 or LANIO_SPEC_DEFS[self.__model].aoNum <= aoNum:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.setAnalogOutput(aoNum, aoRange, val)

    def setAnalogOutputMulti(self, aoNums: List[int], aoRanges: List[LanioAoRange], vals: List[int]) -> bool:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if any([aoNum < 0 or LANIO_SPEC_DEFS[self.__model].aoNum <= aoNum for aoNum in aoNums]) or \
                len(aoNums) > LANIO_SPEC_DEFS[self.__model].aoNum:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.setAnalogOutputMulti(aoNums, aoRanges, vals)

    def getAnalogOutput(self, aoNum: int) -> Tuple[LanioAoRange, int]:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if aoNum < 0 or LANIO_SPEC_DEFS[self.__model].aoNum <= aoNum:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.getAnalogOutput(aoNum)

    def setAnalogOutputVoltage(self, aoNum: int, voltage: float) -> bool:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if aoNum < 0 or LANIO_SPEC_DEFS[self.__model].aoNum <= aoNum:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.setAnalogOutputVoltage(aoNum, voltage)

    def setAnalogOutputCurrent(self, aoNum: int, milliAmpere: float) -> bool:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if aoNum < 0 or LANIO_SPEC_DEFS[self.__model].aoNum <= aoNum:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.setAnalogOutputCurrent(aoNum, milliAmpere)

    def getAnalogOutputByValue(self, aoNum: int) -> Tuple[LanioAoRange, float]:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if aoNum < 0 or LANIO_SPEC_DEFS[self.__model].aoNum <= aoNum:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.getAnalogOutputByValue(aoNum)

    def setInitialAnalogOutput(self, aoNum: int, aoRange: LanioAoRange, val: int) -> bool:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if aoNum < 0 or LANIO_SPEC_DEFS[self.__model].aoNum <= aoNum:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.setInitialAnalogOutput(aoNum, aoRange, val)

    def setInitialAnalogOutputVoltage(self, aoNum: int, voltage: float) -> bool:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if aoNum < 0 or LANIO_SPEC_DEFS[self.__model].aoNum <= aoNum:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.setInitialAnalogOutputVoltage(aoNum, voltage)

    def setInitialAnalogOutputCurrent(self, aoNum: int, milliAmpere: float) -> bool:
        if self.__io is None:
            raise LanioError(LanioErrorType.NOCONNECTED)
        if aoNum < 0 or LANIO_SPEC_DEFS[self.__model].aoNum <= aoNum:
            raise LanioError(LanioErrorType.INVALIDOPERATION)
        return self.__io.setInitialAnalogOutputCurrent(aoNum, milliAmpere)
